;*** Includes
.include "m16def.inc"
.include "mmc.asm"

;*** Constants
.equ	CPUFRQ	=4000000
.equ	BR	=31250
.equ	BAUDRC	=7		; CPUFRQ/((16*BR)-1)
.equ	SECTOR	=113
.equ	T1SCL	=0x05
; Midi Befehle
.EQU	GMreset	=255
.EQU	MIDIalv	=0xFE
.EQU	MIDIclk	=0xF8
.EQU	MIDIstop=0xFC
; Modi
.EQU	smPLAY	=0x01
.EQU	mPLAY	=0
.EQU	smREC	=0x02
.EQU	mREC	=1
.EQU	smFULL	=0x04
.EQU	mFULL	=2
.EQU	smEOS	=0x08
.EQU	mEOS	=3
.EQU	smMMCBS	=0x10
.EQU	mMMCBS	=4
.EQU	smMMCWR	=0x20
.EQU	mMMCWR	=5


;*** Interrupt flag defs
.equ	fOVF0	=0
.equ	sOVF0	=0x01
.equ	fOVF1	=1
.equ	sOVF1	=0x02
.equ	fUTX	=2
.equ	sUTX	=0x04
.equ	fURX	=3
.equ	sURX	=0x08


;*** Addresses in SRAM
.equ	CMD	=0x0060		; MMC command (5 bytes)
.equ	BUF	=0x0200		; MMC data buffer (512 bytes)
.equ	RXBUF	=0x0100		; MIDI RxBuffer


;*** Ports
; PORT D
.equ	RxD	=0		; HW-UART
.equ	TxD	=1		; -"-
.equ	KEYs	=PIND
.equ	KEY1	=2		; Menu key below LCD (input pin)
.equ	KEY2	=3		; -"-
.equ	KEY3	=4		; -"-
.equ	KEY4	=5		; -"-
.equ	KEY5	=6		; -"-
.equ	KEYMASK	=0x7C
; PORT C
.equ	LEDs	=PORTC
; PORT B
.equ	LEDr	=0
.equ	LEDp	=1


;*** REG defs
.def	null	=r1		; 0
.def	oldkey	=r2
.def	work	=r16		; General working register
.def	temp	=r17		; -"-
.def	cnt	=r18		; General counter
.def	inreg	=r3
.def	mode	=r19		; Mode of state machine
.def	intf	=r20		; Interrupt source
.def	t2	=r5		; Timer 1 is used as counter t0, t1
.def	p0	=r6		; Set of midi data and timing
.def	p1	=r7
.def	p2	=r8
.def	sd	=r9
.def	regs	=r10		; Saves SREG
.def	mcntl	=r11		; Bytecounter im Sektor
.def	mcnth	=r12
.def	scntl	=r13		; Sektorcounter in MMC
.def	scnth	=r14
.def	scnts	=r15
.def	dcnt	=r21
.def	cfile	=r22


;***********************************************************
;* 	INT table
;*

.ORG	0
	rjmp	RESET

.ORG	OVF1addr
	rjmp	T1_OVF

.ORG	OVF2addr
	rjmp	T2_OVF

.ORG	URXCaddr
	rjmp	UART_RXCP

.ORG	UTXCaddr
	rjmp	UART_TXCP



;***********************************************************
;* 	INT routines
;*


T2_OVF:	; Timer 2 overflow
	in	regs,	SREG
	sbr	intf,	sOVF0
	out	SREG,	regs
	set
	reti


T1_OVF:	; Timer 1 overflow
	in	regs,	SREG
	inc 	t2
	sbr	intf,	sOVF1
	out	SREG,	regs
	set
	reti


UART_RXCP: ; Received a midi event
	in	regs,	SREG
	in 	inreg, 	UDR		; fetch byte from UART data reg. 
	sbr	intf,	sURX
	sbrc	mode,	mREC
	  rcall	RECMIDI
	out	SREG,	regs
	set
	reti


UART_TXCP: ; Sent a midi event
	in	regs,	SREG
	sbr	intf,	sUTX
	out	SREG,	regs
	set
	reti



;***********************************************************
;* 	The program
;*

RESET:	; Start up
	ldi 	work, 	LOW(RAMEND)	; set up stack
	out 	SPL, 	work
	ldi 	work, 	HIGH(RAMEND)
	out 	SPH, 	work					 

	rcall	INIT_HW
	clr	mode
	clr	intf
	clr	scntl
	clr	scnth
	clr	scnts
	clr	cfile

	ldi	work,	KEYMASK
	mov	oldkey,	work

;	ldi	ZL,	LOW(t_init<<1)
;	ldi	ZH,	HIGH(t_init<<1)
;	rcall	SENDSTRING

;DEBUG
;	sbr 	mode, 	smREC
;	ldi	YH,	HIGH(RXBUF)
;	ldi	ZH,	HIGH(RXBUF)
;	clr	YL
;	clr	ZL
;	in	work,	UDR
;	sbi 	UCSRB, 	RXCIE

;	sbr	mode,	smMMCBS
;	rcall	delay
;	rcall	delay
;	rcall	delay
;	rcall	delay
;	cbr	mode,	smMMCBS

;	t:rjmp t
;/DEBUG

MAIN:
LOOP:	in	work,	KEYs		; get keys
	andi	work,	KEYMASK
	cp	work,	oldkey
	brne	KEY_HDL

  RK:	sbrc	mode,	mPLAY
	  rcall	PLAY1
	sbrc	mode,	mREC
	  rcall	BUF2MMC

	ldi	work,	(1<<SE)
	out	MCUCR,	work
	sbrs	mode,	mPLAY
	  sleep

	rjmp	MAIN


KEY_HDL:mov	oldkey,	work
	cpi	work,	KEYMASK
	breq	MAIN

	sbis	PIND,	2
	  rjmp	PLAY
	sbis	PIND,	4
	  rjmp	REC
	sbis	PIND,	3
	  rjmp	STOP
	sbis	PIND,	5
	  rjmp	INCFN
	sbis	PIND,	6
	  rjmp	DECFN
  RH:	rcall	delay
	rjmp	RK



; *** PLAY Taste

PLAY:	sbrc	mode,	mREC
	  rjmp	RH
	sbrc	mode,	mPLAY
	  rjmp	RH
;	ldi	ZL,	LOW(t_play<<1)
;	ldi	ZH,	HIGH(t_play<<1)
;	rcall	SENDSTRING

	; Modus Play
	sbr 	mode, 	smPLAY
	cbi	PORTB,	LEDp
	; Pointer auf song start
	rcall	INITMIDI
	; Lies ersten Datensatz
	clr	work
	rcall	READ_SECTOR_START
	rcall	READMIDI
	; Timer start
	ldi 	work, 	T1SCL
	out 	TCCR1B,	work
;	ldi	ZL,	LOW(t_tmr<<1)
;	ldi	ZH,	HIGH(t_tmr<<1)
;	rcall	SENDSTRING
	rjmp	RH


; *** REC Taste

REC:	sbrc	mode,	mPLAY
	  rjmp	RH
	sbrc	mode,	mREC
	  rjmp	RH
	; Modus Record
	sbr 	mode, 	smREC
	cbi	PORTB,	LEDr
	; Zhler reset
	rcall	INITMIDI
	ldi	YH,	HIGH(RXBUF)
	ldi	ZH,	HIGH(RXBUF)
	clr	YL
	clr	ZL
	clr	work
	rcall	WRITE_SECTOR_START
	sbr	mode,	smMMCWR
	in	work,	UDR
	sbi 	UCSRB, 	RXCIE
	rjmp	RH


; *** STOP Taste

STOP:	; Stoppt REC und PLAY
	out	TCCR1B,	NULL
	; Wenn im Rec Mode, schreibe STOPP Event und RxInt Disable
	sbrs	mode,	mREC
	  rjmp	ST_
;	ldi	ZL,	LOW(t_stop<<1)
;	ldi	ZH,	HIGH(t_stop<<1)
;	rcall	SENDSTRING

	rcall	WRITESTOP
	cbi 	UCSRB, 	RXCIE
  STP:	ldi	work,	0xFF
	rcall	WCHAR
	cpi	work,	0
	brne	STP	
	cbr	mode,	smMMCWR
	rcall	WRITE_SECTOR_END
  ST_:	sbrc	mode,	mPLAY
	  rcall	READ_SECTOR_END
	sbi	MMC_Write,	MMC_Chip_Select
	clr	mode
	sbi	PORTB,	LEDr
	sbi	PORTB,	LEDp
	rcall	delay
	rcall	INIT_MMC
	; Midigert ausmachen
	rcall	SENDALLOFF

	rcall	INIT_MMC
	rjmp	RH


INCFN:	sbrc	mode,	mREC
	  rjmp	RH
	sbrc	mode,	mPLAY
	  rjmp	RH
	inc	cfile
	rcall	CALCSECTOR
	rjmp	RH


DECFN:	sbrc	mode,	mREC
	  rjmp	RH
	sbrc	mode,	mPLAY
	  rjmp	RH
	dec	cfile
	rcall	CALCSECTOR
	rjmp	RH


CALCSECTOR:; Calculates sector address for current file
	ldi	work,	128
	mul	cfile,	work
	ldi	work,	LOW(SECTOR)
	add	r0,	work
	ldi	work,	HIGH(SECTOR)
	adc	r1,	work
	mov	scntl,	r0
	mov	scnth,	r1
	clr	scnts
	mov	work,	cfile
	com	work
	out	LEDs,	work
	clr	r0
	clr	r1
	ret



; *** PLAY Polling

PLAY1:	sbrs	mode,	mPLAY		; Check Modus Play
	  ret
	sbrc	mode,	mREC		; Check Modus Play
	  ret
	;Timer t2 check
	;rcall	compT1
	;Vergleiche t (Timerstand) mit p (Liedposition)
;	ldi	ZL,	LOW(t_playpoll<<1)
;	ldi	ZH,	HIGH(t_playpoll<<1)
;	rcall	SENDSTRING

	in	work,	TCNT1L
	in	temp,	TCNT1H
	cp	t2,	p2
	breq	NCHK1
	brlo	PEXIT
	rjmp	POSOK
 NCHK1:	cp	temp,	p1
 	breq	NCHK2
	brlo	PEXIT
	rjmp	POSOK
 NCHK2:	cp	work,	p0
	brlo	PEXIT
	; Pos. passt: Gebe Daten wieder, lade neue Daten
 POSOK:	cbr	intf,	sUTX
	clt
	out	UDR,	sd

 WPLAY:	sbrs	intf,	fUTX
	  rjmp	WPLAY
	cbr	intf,	sUTX
	clt

	rcall	READMIDI
 PEXIT:	ret





;***********************************************************
;* 	Additional routines
;*

INIT_HW: ; Inits the micro; no params
	clr 	null

	; Timer setup
	out 	TCCR1B,	NULL		; T1SCL to start
	out	TCNT1H,	NULL
	out 	TCNT1L,	NULL		; Timer 0.5ms
	ldi	work,	0x05
	out	TCCR2,	work
	ldi 	work,	(1<<TOIE1)|(1<<TOIE2)
	out 	TIMSK,	work		; Timer interrupts

	; UART setup
	ldi 	work, 	BAUDRC		; baudrate constant
	out 	UBRRL, 	work		; prepare baud rate register
	sbi 	UCSRB, 	RXEN		; set "receive enable flag"
	sbi 	UCSRB, 	TXEN		; set "transmitter enable flag"
	cbi 	UCSRB, 	RXCIE		; and "received character interrupt enable flag"
	sbi 	UCSRB, 	TXCIE		; and "received character interrupt enable flag"
	sei
	in	inreg,	UDR
	clt
	
	; PORT setup
	ldi	work,	0xBF
 	out	DDRB,	work		; //PB7=SCK,PB6=MISO,PB5=MOSI,PB4=MMC_CS,PB3..0 not used
	ldi	work,	0x13
	out	PORTB,	work
	ldi	work,	KEYMASK
	out	PORTD,	work
	out	DDRD,	null
	ldi	work,	0xFF
	out	DDRC,	work
	out	PORTC,	work

	rcall	INIT_MMC

	ret


SENDSTRING:; Sends a string from ROM over serial conn.
	lpm	cnt,	Z+
	out	UDR,	cnt
  ml1:	brtc	ml1
	clt
	cpi	cnt,	13
	brne	SENDSTRING
	ret


SENDCHAR:; Sends a char in work
	out	UDR,	work
  ml_:	brtc	ml_
	clt
	ret


WRITESTOP:; Schreibt Stoppevent
	ldi	work,	MIDIstop
	mov	inreg,	work
	rcall	WRITEMIDI
	rcall	BUF2MMC
	ret


SENDRESET:; Setzt MIDI Gert zurck
	ldi	work,	GMreset
	out	UDR,	work
	cbr	intf,	sUTX
  WFS:	sbrs	intf,	fUTX
	  rjmp	WFS
	ret


SENDALLOFF:; All off
	ldi	temp,	0x7B
	rcall	SENDALLCH
	ldi	temp,	0x79
	rcall	SENDALLCH
	ret


SENDALLCH:; Sends to all channels
	clr	cnt
 SAO:	ldi	work,	0xB0
	add	work,	cnt
	rcall	SENDCHAR
	mov	work,	temp
	rcall	SENDCHAR
	ldi	work,	0x00
	rcall	SENDCHAR
	inc	cnt
	cpi	cnt,	0x10
	brlo	SAO
	ret


RECMIDI:; Zeichnet ein Midievent auf
;	sbrs	intf,	fURX	; Nur wenn ein Zeichen gekommen ist
;	  ret
	push	XL
	push	XH
	push	work
	push	cnt
	push	temp
	cbr	intf,	sURX
	;Timer t2 check
	;rcall	compT1
	; Prfe Daten
	mov	work,	inreg
	cpi	work,	MIDIclk
	breq	EXR
	cpi	work,	MIDIalv
	breq	EXR
	; Timer start
	cbi	PORTB,	LEDp
	in	work,	TCCR1B
	andi	work,	7
	cpi	work,	T1SCL
	breq	RM1
	ldi 	work, 	T1SCL
	out 	TCCR1B,	work
	; Schreibe Daten
  RM1:	rcall	WRITEMIDI
	
  EXR:	pop	temp
	pop	cnt
	pop	work
	pop	XH
	pop	XL
	ret


READMIDI:; Liest einen Datensatz ein
;	sbrc	mode,	mMMCBS
;	  rjmp	READMIDI
;	ldi	ZL,	LOW(t_getdata<<1)
;	ldi	ZH,	HIGH(t_getdata<<1)
;	rcall	SENDSTRING

	rcall	RDCHAR
	mov	p2,	work
	rcall	RDCHAR
	mov	p1,	work
	rcall	RDCHAR
	mov	p0,	work
	rcall	RDCHAR
	mov	sd,	work
	cpi	work,	MIDIstop
	breq	H_EOS
	sbrc	mode,	mFULL		; Voll?
	  rjmp	H_FULL
	ret
 H_EOS:	; Ende vom Lied
	out	TCCR1B,	NULL
	clr	mode
	; Midigert ausmachen
;	rcall	sendRESET
	rcall	READ_SECTOR_END
	sbi	MMC_Write,	MMC_Chip_Select
	rcall	delay
	rcall	INIT_MMC
	rcall	delay
	rcall	INIT_MMC
	ret


WRITEMIDI:; Speichert Miditiming und Byte
	in	cnt,	TCNT1L
	in	temp,	TCNT1H
	mov 	work, 	t2
	rcall 	WRTCHAR
	mov 	work, 	temp
	rcall 	WRTCHAR
	mov 	work, 	cnt
	rcall 	WRTCHAR
	mov 	work, 	inreg
	rcall 	WRTCHAR
	sbrc	mode,	mFULL		; Voll?
	  rjmp	H_FULL
	ret
 H_FULL:; Speicher fr Song voll!
	out	TCCR1B,	NULL
	clr	mode
	; Midigert ausmachen
;	rcall	sendRESET
	
	ret


WRTCHAR:; Speichert 1 Byte <work>->BUFFER->MMC
	st	Y+,	work
	ldi	YH,	HIGH(RXBUF)
	ret

BUF2MMC:;BUFFER->MMC kopieren
	sbrs	mode,	mMMCWR
	  ret
	sbrc	mode,	mMMCBS
	  ret
	cp	ZL,	YL
	brne	HR
	ret

  HR:	ld	work,	Z+
	ldi	ZH,	HIGH(RXBUF)
	rcall	WCHAR
	cp	ZL,	YL
	brne	HR
	ret

 WCHAR:	rcall	Write_Byte_MMC
	rcall	INCMCNT
	cpi	work,	1
	ldi	work,	0
	brne	EXWC
	cbr	mode,	smMMCWR
	rcall	WRITE_SECTOR_END
	clr	work
	inc	scntl
	brne	NSCW
	inc	scnth
  NSCW:	rcall	WRITE_SECTOR_START
	sbr	mode,	smMMCWR
	ldi	work,	1
  EXWC:	ret
	

RDCHAR:; Liest 1 Byte MMC-><work>
	rcall	Read_Byte_MMC
	push	work
	rcall	INCMCNT
	cpi	work,	1
	brne	EXRC
	rcall	READ_SECTOR_END
	clr	work
	inc	scntl
	brne	NSCT
	inc	scnth
  NSCT:	rcall	READ_SECTOR_START
  EXRC:	pop	work
	ret


INCMCNT:; Erhht den Bytezhler, work=1 falls 512 erreicht! (und Reset Zhler)
	clr	work
	inc	mcntl
	brne	EXC
	inc	mcnth
	mov	temp,	mcnth
	cpi	temp,	2
	brlo	EXC
	clr	mcnth
	clr	mcntl	
	ldi	work,	1
  EXC:	ret


INITMIDI: ; Midi Rekorder/Player vorbereiten
	; Zhler reset
	clr 	t2
	cbr	mode,	smFULL
	cbr	mode,	smEOS
	; Timer 1 Reset
	out	TCCR1B,	NULL
	out	TCNT1H,	NULL
	out	TCNT1L,	NULL
	; Ldt Seiten-Adresse
	rcall	CALCSECTOR
	clr	mcntl
	clr	mcnth
	ret


;t_init:
;.DB	"init ok",13

;t_play:
;.DB	"play pressed",13

;t_playpoll:
;.DB	"play polling",13

;t_tmr:
;.DB	"timer running",13

;t_stop:
;.DB	"stop pressed",13

;t_getdata:
;.DB	"getting data from MMC",13
